<?php
// $Id: pifr.cron.inc,v 1.11 2008/11/18 05:03:40 boombatower Exp $
/**
 * @file
 * Automatically review patches attached to issues.
 *
 * Copyright 2008 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Check if cron should run for the specified role and if cron should run then
 * updates the last run variable to the current time.
 *
 * @param string $role Role string: test_master, or project.
 * @return boolean Cron should run.
 */
function pifr_cron_check($role) {
  $last = variable_get('pifr_' . $role . '_cron_last', 0);
  $delay = variable_get('pifr_' . $role . '_cron_delay', 5);
  $run = ($last + ($delay * 60) <= time());

  if ($run) {
    variable_set('pifr_' . $role . '_cron_last', time());
  }
  return $run;
}

/**
 * Find any free slaves and send them a file to review.
 */
function pifr_send_free_slaves() {
  $slaves = pifr_server_free_slaves();
  foreach ($slaves as $slave) {
    // Free slave found, send next file in queue.
    $file = pifr_file_next();
    if ($file) {
      pifr_send_to_slave($slave, $file);
    }
    else {
      break; // No files in queue.
    }
  }
}

/**
 * Send the results that have not yet been sent.
 */
function pifr_send_results() {
  $servers = pifr_server_get_type(PIFR_SERVER_PROJECT);
  $last = variable_get('pifr_result_last', 0);
  variable_set('pifr_result_last', time());
  $self = pifr_server_get_type(PIFR_SERVER_TEST_MASTER);

  foreach ($servers as $server) {
    // XXX PIFT specific server array.
    $server_param = array('url' => url('', array('absolute' => TRUE)) . 'xmlrpc.php', 'server_key' => $self->server_key);

    $batches = pifr_get_results($server->server_id, $last);

    foreach ($batches as $batch) {
      // XXX PIFT specific.
      $response = xmlrpc($server->url . 'xmlrpc.php', 'pift.test.results', $server_param, $batch);
//      $response = xmlrpc($server->url . 'xmlrpc.php', 'pifr.batch.result', $self->server_key, $batch);

      if ($response === FALSE) {
        watchdog('pifr', 'Failed to send batch of results to project server @server_id: @message',
                          array('@server_id' => $server->server_id, '@message' => xmlrpc_error_msg()), WATCHDOG_ERROR);
      }
      else if (!empty($response)) {
//      else if ($response != PIFR_RESPONSE_ACCEPTED) {
        watchdog('pifr', 'Failed to send batch of results to project server @server_id.',
                          array('@server_id' => $server->server_id), WATCHDOG_ERROR);
      }
    }
  }
}

/**
 * Get the results that have not yet been sent to the specified server.
 *
 * @param integer $server_id Server ID.
 * @param integer $last Last timestamp results were sent.
 * @return array List of batches to be sent.
 */
function pifr_get_results($server_id, $last) {
  require_once drupal_get_path('module', 'pifr') . '/pifr.pages.inc';

  $batches = array();
  $batch = array();

  // Get results since last file ID sent.
  $results = db_query('SELECT *
                       FROM {pifr_file}
                       WHERE server_id = %d
                       AND last_tested > %d
                       AND status > %d
                       ORDER BY server_id, nid, cid', $server_id, $last, PIFR_FILE_ERROR);

  while ($result = db_fetch_object($results)) {
    if (count($batch) == PIFR_MAX_FILE_COUNT) {
      $batches[] = $batch;
      $batch = array();
    }

    $file = array();
    $file['server_file_id'] = $result->server_file_id;
    $file['nid'] = $result->nid;
    $file['cid'] = $result->cid;
    $file['filename'] = $result->filename;
    $file['pass'] = ($result->status == PIFR_FILE_TEST_PASS);
    $file['summary'] = pifr_file_get_result($result) . ' ' .
                       l('Detailed results', url('pifr/file/' . $server_id . '/' . $result->filename, array('absolute' => TRUE)));

    $batch[] = pifr_convert_batch_result_file($file);
  }

  if (count($batch)) {
    $batches[] = $batch;
  }

  return $batches;
}

/**
 * Convert PIFR format.
 *
 * @param array $file File data result array in PIFR format.
 * @return array File result data array.
 */
function pifr_convert_batch_result_file($file) {
  $file['ftid'] = $file['server_file_id'];
  $file['status'] = ($file['pass'] == PIFR_FILE_TEST_PASS ? 1 : 2);
  $file['display_data'] = $file['summary'];

  unset($file['server_file_id']);
  unset($file['nid']);
  unset($file['cid']);
  unset($file['filename']);
  unset($file['pass']);
  unset($file['summary']);
  return $file;
}

/**
 * Ensure that drupal.org is running and that Drupal HEAD passes tests.
 */
function pifr_check_project_server() {
  $response = drupal_http_request('http://drupal.org/cvs?rss=true&nid=3060&branch=HEAD');
  $up = $response->code == 200;

  drupal_set_message('pifr_check_project_server() -> ' . $response->code);
  if ($up) {
    // Check commit feed for new commits.
    $xml = simplexml_load_string($response->data);

    if (isset($xml->channel->item[0]) && preg_match('/Commit (\d+) by (\w+)/', (string) $xml->channel->item[0]->title, $match)) {
      if (variable_get('pifr_commit_last', '') != $match[1]) {
        // New commit to Drupal HEAD, schedule blank patch test.
        $file = pifr_get_empty_patch();

        // Ensure that the blank file is tested next.
        pifr_file_set_queued($file_id);
        db_query('UPDATE {pifr_file} SET test_count = %d WHERE file_id = %d', -1, $file_id);

        variable_set('pifr_commit_last', $match[1]);
        variable_set('pifr_commit_last_author', $match[2]);
      }
    }
    else {
      watchdog('pifr', 'Failed to parse commit feed.', array(), WATCHDOG_ERROR);
      $up = FALSE;
    }
  }
  else {
    watchdog('pifr', 'Drupal.org is not responding.', array(), WATCHDOG_ERROR);
  }

  // If testing master server is not active then force test HEAD.
  if (!variable_get('pifr_test_master_active', FALSE) && $up) {
    $file = pifr_get_empty_patch();
    if ($file->status == PIFR_FILE_QUEUED) {
      pifr_send_to_slave($slave, $file);
    }
  }

  pifr_set_server_status($up);
}

/**
 * Get the empty patch file record.
 *
 * @return object File object of empty record.
 */
function pifr_get_empty_patch() {
  $server = pifr_server_find('drupal.org', PIFR_SERVER_PROJECT);

  if (!($file = pifr_file_find(PIFR_BLANK_PATCH, $server->server_id))) {
    $file_id = pifr_file_insert($server->server_id, PIFR_BLANK_FILE_ID, -1, -1, PIFR_BLANK_PATCH);
    $file = pifr_file_get($file_id);
  }
  return $file;
}

/**
 * Set the testing master server status.
 *
 * @param boolean $active Server active.
 * @param boolean $commit_fail Failure caused by commit.
 */
function pifr_set_server_status($active, $commit_fail = FALSE) {
  global $language;

  if ($active) {
    // Drupal.org is running and HEAD passes tests, continue testing or
    // reactivate if drupal.org went down temporarily.
    variable_set('pifr_test_master_active', TRUE);
    variable_set('pifr_test_master_report', TRUE);
  }
  else {
    // Deactivate testing and inform developers via dev list.
    variable_set('pifr_test_master_active', FALSE);
    variable_set('pifr_test_master_report', FALSE);

    if ($commit_fail) {
      drupal_mail('pifr', 'dev_alert', 'development@drupal.org', $language, array(), 'noreply@testing.drupal.org');
    }
    else {
      drupal_set_message('not active');
      $result = db_query("SELECT u.uid
                          FROM {users} u, {users_roles} ur, {role} r
                          WHERE u.uid = ur.uid
                          AND ur.rid = r.rid
                          AND r.name = '%s'", 'admin');
      while ($uid = db_fetch_object($result)) {
        $user = user_load(array('uid' => $uid));
        drupal_set_message('user: ' . $user->mail);
        drupal_mail('pifr', 'admin_alert', $user->mail, $language, array(), 'noreply@testing.drupal.org');
      }
    }
  }
}

/**
 * Implementation of hook_mail().
 */
function pifr_mail($key, &$message, $params) {
  $language = $message['language'];

  if ($key == 'dev_alert') {
    $message['subject'] = t('Drupal 7 HEAD is broken!', array(), $language->language);
    $message['body'] = t("Dev list,

ALERT! HEAD was broken in commit #@commit (@url) by @author.

   ..-^~~~^-..
 .~           ~.
(;:           :;)
 (:           :)
   ':._   _.:'
       | |
     (=====)
       | |
       | |
       | |
    ((/   \))

I have decided to take my leave at this opportune time, thus testing will be
disabled until this issue is fixed.

Once HEAD has been fix testing will automatically resume.

Thanks,

The humble test bot
http://drupal.org/user/180064", array('@url' => 'http://drupal.org/cvs?commit=' . variable_get('pifr_commit_last', ''),
                                      '@commit' => variable_get('pifr_commit_last', ''),
                                      '@author' => variable_get('pifr_commit_last_author', '')), $language->language);
  }
  else if ($key == 'admin_alert') {
    $message['subject'] = t('Testing has been temporarily shut off', array(), $language->language);
    $message['body'] = "Testing administrator,

Drupal.org is unreachable and thus testing has been temporarily shut off. Once
drupal.org becomes reachable the server will automatically resume testing.

Thanks,

The humble test bot
http://drupal.org/user/180064";
  }
}
