<?php
// $Id: pifr.review.inc,v 1.16 2008/11/04 06:33:19 boombatower Exp $
/**
 * @file
 * Automatically review patches attached to issues.
 *
 * Copyright 2008 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Start file review process.
 *
 * @param string $server_url Server URL sending file.
 * @param integer $file_id File ID.
 * @param array $filepath The path to file on project site.
 */
function pifr_review_start($server_url, $file_id, $filepath) {
  // Store variables to indicate status.
  variable_set('pifr_reviewing', TRUE);
  variable_set('pifr_server', $server_url);
  variable_set('pifr_file_id', $file_id);
  variable_set('pifr_file', $filepath);

  // Run review in background so testing master server can close connection.
  pifr_review_exec_background('curl ' . url('pifr/review', array('absolute' => TRUE)));
}

/**
 * Run review process.
 */
function pifr_review_run() {
  if (!variable_get('pifr_reviewing', FALSE)) {
    // Possible invalid attempt to run slave.
    return;
  }

  $file_id = variable_get('pifr_file_id', '');
  $filepath = variable_get('pifr_file', '');

  // Fetch file and save on local server.
  if (!pifr_review_fetch($filepath)) {
    // Inform master server that file failed to download.
    pifr_review_send($file_id, PIFR_FILE_ERROR, 'Failed to fetch file.');
    return;
  }

  // File downloaded, checkout latest CVS.
  if (!pifr_review_checkout()) {
    // Inform master server that HEAD failed to checkout.
    pifr_review_send($file_id, PIFR_FILE_ERROR, 'Failed to checkout HEAD.');
    return;
  }

  // Apply patch to latest CVS.
  if (!pifr_review_apply_patch(basename($filepath))) {
    // Inform master server that patch does not apply.
    pifr_review_send($file_id, PIFR_FILE_TEST_FAIL, 'Failed to apply patch.');
    return;
  }

  // Check syntax of files changed by patch.
  if (!pifr_review_check_syntax(basename($filepath))) {
    // Inform master server that patch create invalid PHP syntax.
    pifr_review_send($file_id, PIFR_FILE_TEST_FAIL, 'Invalid PHP syntax.');
    return;
  }

  // Install HEAD.
  if (!pifr_review_install()) {
    // Inform master server that could not install HEAD.
    pifr_review_send($file_id, PIFR_FILE_TEST_FAIL, 'Failed to install HEAD.');
    return;
  }

  // Run tests.
  if (!pifr_review_run_tests()) {
    // Inform master server that could not run tests.
    pifr_review_send($file_id, PIFR_FILE_TEST_FAIL, 'Failed to run tests.');
    return;
  }

  // Send testing results as XML. Server will evaluate the PASS based on the summary XML.
  pifr_review_send($file_id, PIFR_FILE_TEST_PASS, '', $GLOBALS['pifr_xml_results']);
}

/**
 * Fetch the file to review and save locally on server.
 *
 * @param string $filepath Path to file on project server.
 * @return boolean Success.
 */
function pifr_review_fetch($filepath) {
  $response = drupal_http_request($filepath);
  $directory = file_directory_path() . '/review';
  if ($response->code == 200 && file_check_directory($directory, TRUE)) {
    $path = $directory . '/' . basename($filepath);
    file_save_data($response->data, $path,  FILE_EXISTS_REPLACE);
    return file_exists($path);
  }
  return FALSE;
}

/**
 * Clear the checkout directory and checkout the latest Drupal 7 HEAD. This
 * assumes a writable checkout directory.
 *
 * @return boolean Success.
 */
function pifr_review_checkout() {
  // Use checkout directory for convenience.
  pifr_review_working_directory();

  // Clear directory.
  if (file_exists('./checkout')) {
    // Change permissions on settings.php which is read-only after installation.
    chmod('./checkout/sites/default', 0777);

    exec('rm -r ./checkout', $output, $status);
    if ($status != 0) {
      watchdog('pifr', 'Failed to clear checkout directory.', array(), WATCHDOG_ERROR);
    }
  }

  // Re-create directory.
  mkdir('./checkout');

  // Checkout latest Drupal 7 HEAD.
  $cvs_checkout = 'env - cvs -z6 -Q -d:pserver:anonymous:anonymous@cvs.drupal.org:/cvs/drupal co -dcheckout -r HEAD drupal';
  exec($cvs_checkout, $output, $status);

  // Revert to original working directory.
  pifr_review_working_directory(TRUE);
  return !$status;
}

/**
 * Apply the patch to the lastest code from CVS.
 *
 * @param string $filename Patch filename.
 * @return boolean Success.
 */
function pifr_review_apply_patch($filename) {
  pifr_review_working_directory();
  chdir('./checkout');

  exec('patch -p0 < ../review/' . $filename, $output, $status);

  pifr_review_working_directory(TRUE);
  return !$status;
}

/**
 * Check the syntax of the code changed by the patch.
 *
 * @param string $filename Patch filename.
 * @return boolean Success.
 */
function pifr_review_check_syntax($filename) {
  pifr_review_working_directory();

  // Find all files changed by patch.
  $contents = file_get_contents('./review/' . $filename);
  preg_match_all('/^Index: (.*)$/m', $contents, $matches, PREG_SET_ORDER);

  chdir('./checkout');
  $valid = TRUE;
  foreach ($matches as $match) {
    exec('php -l -f ./' . $match[1], $output, $status);

    if ($status != 0) {
      $valid = FALSE;
      break;
    }
  }

  pifr_review_working_directory(TRUE);
  return $valid;
}

/**
 * Install Drupal HEAD.
 *
 * @return boolean Success.
 */
function pifr_review_install() {
  global $db_url;

  // Ensure we can run site by using symbolic link.
  if (!file_exists('./checkout')) {
    watchdog('pifr', 'Failed to find symbolic link needed during install process, see INSTALL.txt.', array(), WATCHDOG_ERROR);
    return FALSE;
  }

  // Get database information.
  $url = parse_url($db_url);
  $db = substr(urldecode($url['path']), 1) . '_checkout';
  $user = urldecode($url['user']);
  $pass = urldecode($url['pass']);

  // Clear database of previous instance and any left of testing tables.
  exec('mysql -u' . $user . ' -p' . $pass . ' ' . $db . ' -e "SHOW TABLES"' .
       ' | grep -v Tables_in | grep -v "+"' .
       ' | \gawk \'{print "DROP TABLE " $1 ";"}\'' .
       ' | mysql -u' . $user . ' -p' . $pass . ' ' . $db, $output, $status);
  if ($status != 0) {
    watchdog('pifr', 'Failed to clear test database.', array(), WATCHDOG_ERROR);
  }

  // Make a copy of the settings file to ensure there is no permission issue.
  pifr_review_working_directory();
  $default_settings = './checkout/sites/default/default.settings.php';
  $settings = './checkout/sites/default/settings.php';
  file_copy($default_settings, $settings);
  pifr_review_working_directory(TRUE);

  require_once drupal_get_path('module', 'simpletest') . '/drupal_web_test_case.php';
  pifr_review_path();

  // Install Drupal.
  $d = new DrupalWebTestCase();

  // Step: Select an installation profile.
  // Step: Choose language.
  $d->drupalGet('install.php', array('query' => 'profile=default&locale=en'));

  // Step: Database configuration.
  $edit = array();
  $edit['database'] = $db;
  $edit['username'] = $user;
  $edit['password'] = $pass;
  $edit['db_prefix'] = '';
  $d->drupalPost(NULL, $edit, t('Save and continue'));

  // Step: Site configuration.
  $edit = array();
  $edit['site_name'] = 'checkout';
  $edit['site_mail'] = 'admin@example.com';
  $edit['account[name]'] = 'admin';
  $edit['account[mail]'] = 'admin@example.com';
  $edit['account[pass][pass1]'] = $pass = $d->randomName(12);
  $edit['account[pass][pass2]'] = $pass;
  $edit['clean_url'] = 0;
  $edit['update_status_module[1]'] = FALSE;
  $d->drupalPost(NULL, $edit, t('Save and continue'));

  // Record password use later.
  pifr_review_admin_pass($pass);

  // Step: Finished.
  $d->assertText(t('Drupal installation complete'));

  pifr_review_path(TRUE);

  // Clear SimpleTest results.
  simpletest_clean_results_table();

  // Make sure that site installed correctly.
  return ($d->_results['#exception'] + $d->_results['#fail']) == 0;
}

/**
 * Run tests on Drupal HEAD.
 *
 * @return boolean Success.
 */
function pifr_review_run_tests() {
  global $pifr_xml_results;

  // Login to Drupal HEAD and enable the SimpleTest module.
  pifr_review_path();
  $pass = pifr_review_admin_pass();

  $d = new DrupalWebTestCase();

  // Login as admin.
  $user = new stdClass();
  $user->name = 'admin';
  $user->pass_raw = $pass;
  $d->drupalLogin($user);

  // Enable SimpleTest.
  $edit = array();
  $edit['modules[Core][simpletest][enable]'] = TRUE;
  $d->drupalPost('admin/build/modules', $edit, t('Save configuration'));

  // Run tests from command line.
  pifr_review_working_directory();
  chdir('./checkout');

  $php = variable_get('pifr_php_executable', '/usr/bin/php');
  $url = escapeshellarg(url('', array('absolute' => TRUE)));
  $results = shell_exec($php . ' ./scripts/run-tests.sh --php ' . $php . ' --url ' . $url . ' --all');
  $pifr_xml_results = pifr_review_format_results($results);

  pifr_review_working_directory(TRUE);
  pifr_review_path(TRUE);

  // Clear SimpleTest results.
  simpletest_clean_results_table();

  return ($pifr_xml_results && ($d->_results['#exception'] + $d->_results['#fail']) == 0);
}

/**
 * Format the results from the command line return into a compressed XML dump
 * that can be sent back to testing master.
 *
 * @param string $results Raw command line output.
 * @return binary Compressed XML results.
 */
function pifr_review_format_results($results) {
  if (preg_match_all('/^(.*?) (\d+) passes, (\d+) fails, and (\d+) exceptions$/m', $results, $matches, PREG_SET_ORDER)) {
    $xml = new SimpleXMLElement('<results />');
    $classes = $xml->addChild('classes');
    foreach ($matches as $match) {
      $class = $classes->addChild('class');

      // Fill in class element.
      $class->addChild('name', $match[1]);
      $class->addChild('pass', $match[2]);
      $class->addChild('fail', $match[3]);
      $class->addChild('exception', $match[4]);
    }
    return gzdeflate($xml->asXML(), 9);
  }
  return FALSE;
}

/**
 * Cleanup after reviewing file and set the slave status to free.
 */
function pifr_review_cleanup() {
  // Remove downloaded patch.
  $file = file_directory_path() . '/review/' . basename(variable_get('pifr_file', ''));
  if (is_file($file)) {
    unlink($file);
  }

  variable_set('pifr_reviewing', FALSE);
  variable_del('pifr_server');
  variable_del('pifr_file_id');
  variable_del('pifr_file');
}

/**
 * Send file review to testing master.
 *
 * @param interger $file_id File ID.
 * @param interger $status Review status.
 * @param string $return String message.
 * @param string $results Test results exported as XML and compressed.
 */
function pifr_review_send($file_id, $status, $return, $results = NULL) {
  if ($status == PIFR_FILE_ERROR) {
    watchdog('pifr', $return, array(), WATCHDOG_ERROR);
  }

  $server_url = variable_get('pifr_server', '') . 'xmlrpc.php';
  $self = pifr_server_get_type(PIFR_SERVER_SLAVE);

  // Ensure that slave review environment is reset, so that it can accept another file.
  pifr_review_cleanup();

  require_once 'includes/xmlrpc.inc';
  $response = xmlrpc($server_url, 'pifr.file.result', $self[0]->server_key, (int) $file_id, (int) $status, $return, xmlrpc_base64($results));
}

/**
 * Set working directory to original, Drupal set, or files directory.
 *
 * @param boolean $original Original directory.
 */
function pifr_review_working_directory($original = FALSE) {
  static $working_directory;

  if (!isset($working_directory)) {
    $working_directory = getcwd();
  }

  if ($original) {
    chdir($working_directory);
  }
  else {
    chdir(file_directory_path());
  }
}

/**
 * Set the original, Drupal set, or base_url and base_path to /checkout.
 *
 * @param boolean $original Original path.
 */
function pifr_review_path($original = FALSE) {
  global $base_url, $base_path;
  static $base_url_original, $base_path_original;

  if (!isset($base_url_original) || !isset($base_path_original)) {
    $base_url_original = $base_url;
    $base_path_original = $base_path;
  }

  if ($original) {
    $base_url = $base_url_original;
    $base_path = $base_path_original;
  }
  else {
    $base_url .= '/checkout';
    $base_path = '/checkout';
  }
}

/**
 * Store or retreive password.
 *
 * @param string $password Password to store.
 * @return string Stored password.
 */
function pifr_review_admin_pass($password = NULL) {
  static $_password;

  if ($password) {
    $_password = $password;
  }
  return $_password;
}

/**
 * Execute a system command in the background.
 */
function pifr_review_exec_background($command) {
  return shell_exec("nohup $command > /dev/null 2> /dev/null & echo $!");
}
